/*	$NetBSD: aica.c,v 1.18 2008/08/01 20:19:49 marcus Exp $	*/

/*
 * Copyright (c) 2003 SHIMIZU Ryo <ryo@misakimix.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$NetBSD: aica.c,v 1.18 2008/08/01 20:19:49 marcus Exp $");

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/device.h>
#include <sys/kernel.h>
#include <sys/proc.h>
#include <sys/syslog.h>
#include <sys/audioio.h>
#include <dev/audio_if.h>
#include <dev/mulaw.h>
#include <dev/auconv.h>
#include <machine/bus.h>
#include <machine/sysasicvar.h>
#include <dreamcast/dev/g2/g2busvar.h>
#include <dreamcast/dev/g2/aicareg.h>
#include <dreamcast/dev/g2/aicavar.h>
#include <dreamcast/dev/microcode/aica_armcode.h>

#define AICA_NCHAN	64
#define AICA_NAUDIO	4	/* aica privides multi audio devices */

struct aica_softc {
	struct device sc_dev;
	bus_space_tag_t sc_memt;
	bus_space_handle_t sc_aica_regh;
	bus_space_handle_t sc_aica_memh;
	void *sc_ih;
	struct audio_encoding_set *sc_encodings;

	/* audio property */
	int sc_aica_enabled;
	int sc_buffersize;
	int sc_output_master;

	/* handling AICA */
	struct cmdqueue sc_cmdq_to_aica;
	struct cmdqueue sc_cmdq_from_aica;
	struct cmd sc_cmdbuf;
	uint32_t sc_msgpos;
	uint8_t sc_msgbuf[AICA_MSGBUFSIZE + 1];

	/* for each audio devices */
	struct aica_audioarg {
		struct aica_softc *aaa_sc;
		int aaa_open;
		int aaa_unit;
		int aaa_playing;

		int aaa_gain_left;
		int aaa_gain_right;

		int aaa_precision;
		int aaa_channels;
		int aaa_rate;
		int aaa_blksize_per_ch;
		int aaa_nblk;
		int aaa_nextblk;
		int aaa_usedblk;
		int aaa_nsamples;

		/* audio callback */
		void (*aaa_intr)(void *);
		void *aaa_intr_arg;
	} sc_audioarg[AICA_NAUDIO];
};

static const struct audio_format aica_formats[] = {
	{
		NULL, AUMODE_PLAY, AUDIO_ENCODING_ADPCM, 4, 4,
		1, AUFMT_MONAURAL, 0, {60, 65535}
	},
	{
		NULL, AUMODE_PLAY, AUDIO_ENCODING_SLINEAR_LE, 8, 8,
		1, AUFMT_MONAURAL, 0, {60, 65535}
	},
	{
		NULL, AUMODE_PLAY, AUDIO_ENCODING_SLINEAR_LE, 8, 8,
		2, AUFMT_STEREO, 0, {60, 65535}
	},
	{
		NULL, AUMODE_PLAY, AUDIO_ENCODING_SLINEAR_LE, 16, 16,
		1, AUFMT_MONAURAL, 0, {60, 65535}
	},
	{
		NULL, AUMODE_PLAY, AUDIO_ENCODING_SLINEAR_LE, 16, 16,
		2, AUFMT_STEREO, 0, {60, 65535}
	},
};

int aica_match(struct device *, struct cfdata *, void *);
void aica_attach(struct device *, struct device *, void *);
int aica_print(void *, const char *);

CFATTACH_DECL(aica, sizeof(struct aica_softc), aica_match, aica_attach,
    NULL, NULL);

const struct audio_device aica_device = {
	"Dreamcast Sound",
	"",
	"aica"
};

void aica_init(struct aica_softc *);
void aica_initqueue(struct aica_softc *);
void aica_load_microcode(struct aica_softc *, void *, uint32_t);
void aica_heartbeat(struct aica_softc *);
void aica_enable(struct aica_softc *);
void aica_disable(struct aica_softc *);
void aica_memread(struct aica_softc *, bus_size_t, uint32_t *, int);
void aica_memwrite(struct aica_softc *, bus_size_t, const uint32_t *, int);
void aica_ch2p16write(struct aica_softc *, bus_size_t, uint16_t *, int);
void aica_ch2p8write(struct aica_softc *, bus_size_t, uint8_t *, int);
void aica_blockdone(struct aica_audioarg *);
void aica_fillbuffer(struct aica_softc *, struct aica_audioarg *, void *);
int aica_blocksize(struct aica_softc *, int);
void aica_proceed(struct aica_softc *);
int aica_intr(void *);

void aica_set_mvol(struct aica_softc *, int);
void aica_set_vol(struct aica_softc *, int, int, int);

struct cmd *aica_readcmd(struct aica_softc *, struct cmdqueue *);
void aica_sendcmd(struct aica_softc *, struct cmdqueue *, struct cmd *);
void aica_sendcmd_channel_rate(struct aica_softc *, int, int, int);
void aica_sendcmd_prec_blksize(struct aica_softc *, int, int, int);
void aica_sendcmd_nblk_nsample(struct aica_softc *, int, int, int);
void aica_sendcmd_play(struct aica_softc *, int);
void aica_sendcmd_stop(struct aica_softc *, int);
void aica_msgbuf(struct aica_softc *);

/* for audio */
int aica_open(void *, int);
void aica_close(void *);
int aica_query_encoding(void *, struct audio_encoding *);
int aica_set_params(void *, int, int, audio_params_t *, audio_params_t *,
                    stream_filter_list_t *, stream_filter_list_t *);
int aica_round_blocksize(void *, int, int, const audio_params_t *);
int aica_init_output(void *, void *, int);
int aica_init_input(void *, void *, int);
int aica_start_output(void *, void *, int , void (*)(void*), void *);
int aica_start_input(void *, void *, int , void (*)(void*), void *);
int aica_halt_output(void *);
int aica_halt_input(void *);
int aica_getdev(void *, struct audio_device *);
int aica_set_port(void *, mixer_ctrl_t *);
int aica_get_port(void *, mixer_ctrl_t *);
int aica_query_devinfo(void *, mixer_devinfo_t *);
int aica_get_props(void *);

const struct audio_hw_if aica_hw_if = {
	.open			= aica_open,
	.close			= aica_close,
	.drain			= NULL,
	.query_encoding		= aica_query_encoding,
	.set_params		= aica_set_params,
	.round_blocksize	= aica_round_blocksize,
	.commit_settings	= NULL,
	.init_output		= aica_init_output,
	.init_input		= aica_init_input,
	.start_output		= aica_start_output,
	.start_input		= aica_start_input,
	.halt_output		= aica_halt_output,
	.halt_input		= aica_halt_input,
	.speaker_ctl		= NULL,
	.getdev			= aica_getdev,
	.setfd			= NULL,
	.set_port		= aica_set_port,
	.get_port		= aica_get_port,
	.query_devinfo		= aica_query_devinfo,
	.allocm			= NULL,
	.freem			= NULL,
	.round_buffersize	= NULL,
	.mappage		= NULL,
	.get_props		= aica_get_props,
	.trigger_output		= NULL,
	.trigger_input		= NULL,
	.dev_ioctl		= NULL,
	.powerstate		= NULL
};

static void
xxxdump(void *buf, int len)
{
	int i;
	uint8_t *p;
	int col;

	p = buf;
	for (col = 0, i = 0; i < len; i++) {
		if (col == 0)
			log(LOG_DEBUG, "%04x:", i);

		addlog(" %02x", p[i]);
		col++;

		if (col >= 16) {
			addlog("\n");
			col = 0;
		}
	}
	if (col != 0)
		addlog("\n");
}

void
aica_heartbeat(struct aica_softc *sc __unused)
{
#if 1
	uint32_t v;

	v = bus_space_read_4(sc->sc_memt, sc->sc_aica_memh, AICA_HEARTBEAT);
	log(LOG_DEBUG, "HEARTBEART=0x%08x", v);

	v = bus_space_read_4(sc->sc_memt, sc->sc_aica_memh, AICA_DEBUGPORT1);
	addlog(", DEBUG1=0x%08x", v);

	v = bus_space_read_4(sc->sc_memt, sc->sc_aica_memh, AICA_DEBUGPORT2);
	addlog(", DEBUG2=0x%08x\n", v);
#endif
}

/* ARGSUSED */
int
aica_match(struct device *parent __unused, struct cfdata *cf __unused,
           void *aux __unused)
{
	static int aica_matched = 0;

	if (aica_matched)
		return 0;

	aica_matched = 1;
	return 1;
}

/* ARGSUSED */
void
aica_attach(struct device *parent __unused, device_t self, void *aux)
{
	struct aica_softc *sc;
	struct g2bus_attach_args *ga;
	const char *intrstr;
	uint32_t n;
	int rc;

	sc = device_private(self);
	ga = aux;
	sc->sc_memt = ga->ga_memt;

	aprint_naive(": Audio Controller\n");
	aprint_normal(": ARM7DI Sound Processing Unit\n");

	if (bus_space_map(sc->sc_memt, AICA_REG_ADDR_SH4, AICA_REG_SIZE, 0,
	    &sc->sc_aica_regh) != 0) {
		aprint_error_dev(&sc->sc_dev,
		    "couldn't map AICA register space\n");
		goto attach_failure_regmap;
	}
	if (bus_space_map(sc->sc_memt, AICA_RAM_START_SH4, AICA_RAM_SIZE, 0,
	    &sc->sc_aica_memh) != 0) {
		aprint_error_dev(&sc->sc_dev,
		    "couldn't map AICA memory space\n");
		goto attach_failure_memmap;
	}


	/* detect AICA version */
	n = bus_space_read_4(sc->sc_memt, sc->sc_aica_regh, AICA_VOLUME_REG);
	aprint_normal_dev(&sc->sc_dev,
	    "%dbit DAC, %dMB Memory, AICA Version %d\n",
	    n & 0x20 ? 18 : 16,
	    n & 0x10 ? 8 : 2,
	    (n >> 4) & 0x0f);

	aica_disable(sc);	/* Stop AICA ARM7 CPU */
	aica_load_microcode(sc, aica_armcode, sizeof(aica_armcode));
	aica_init(sc);

	/* clear AICA interrupt */
	bus_space_write_4(sc->sc_memt, sc->sc_aica_regh,
	    AICA_INTRST_REG, AICA_INT_CPU);

	/* set interrupt handler */
	intrstr = sysasic_intr_string(SYSASIC_IRL9);
	sc->sc_ih = sysasic_intr_establish(SYSASIC_EVENT_AICA, IPL_AUDIO,
	    SYSASIC_IRL9, aica_intr, sc);
	if (sc->sc_ih == NULL) {
		aprint_error_dev(&sc->sc_dev, "couldn't establish interrupt");
		if (intrstr != NULL)
			aprint_error(" at %s", intrstr);
		aprint_error("\n");
		goto attach_failure;
	}
	aprint_normal_dev(&sc->sc_dev, "interrupting at %s\n", intrstr);

	/* setup audio encodings */
	if ((rc = auconv_create_encodings(aica_formats,
	    __arraycount(aica_formats),  &sc->sc_encodings)) != 0) {
		aprint_error_dev(self,
		    "auconv_create_encoding: error=%d\n", rc);
		goto attach_failure;
	}

	/* attach audio0, audio1, ... */
	for (n = 0; n < AICA_NAUDIO; n++) {
		sc->sc_audioarg[n].aaa_sc = sc;
		sc->sc_audioarg[n].aaa_unit = n;

		audio_attach_mi(&aica_hw_if, &sc->sc_audioarg[n], &sc->sc_dev);
	}

	return;

 attach_failure:
	bus_space_unmap(sc->sc_memt, sc->sc_aica_memh, AICA_RAM_SIZE);
 attach_failure_memmap:
	bus_space_unmap(sc->sc_memt, sc->sc_aica_regh, AICA_REG_SIZE);
 attach_failure_regmap:
	return;
}

void
aica_initqueue(struct aica_softc *sc)
{
	/* init command queue to/from AICA */
	sc->sc_cmdq_to_aica.address = AICA_SH4toARM_CMDQBUF;
	sc->sc_cmdq_to_aica.size = AICA_SH4toARM_CMDQSIZE;
	sc->sc_cmdq_to_aica.index = 0;
	sc->sc_cmdq_to_aica.serial = 0;

	sc->sc_cmdq_from_aica.address = AICA_ARMtoSH4_CMDQBUF;
	sc->sc_cmdq_from_aica.size = AICA_ARMtoSH4_CMDQSIZE;
	sc->sc_cmdq_from_aica.index = 0;
	sc->sc_cmdq_from_aica.serial = 0;
}

void
aica_init(struct aica_softc *sc)
{
	int ch, n;

	aica_set_mvol(sc, 0);

	for (ch = 0; ch < AICA_NCHAN; ch++) {
		bus_space_write_4(sc->sc_memt,sc->sc_aica_regh,
		    AICA_PLAYCTRL_REG(ch),
		    AICA_PLAYCTRL_KEYOFF |
		    AICA_PLAYCTRL_LOOP |
		    AICA_PLAYCTRL_PCM16BIT);

		bus_space_write_4(sc->sc_memt,sc->sc_aica_regh,
		    AICA_ADDRLO_REG(ch), 0);
		bus_space_write_4(sc->sc_memt,sc->sc_aica_regh,
		    AICA_LOOP_START_REG(ch), 0);
		bus_space_write_4(sc->sc_memt,sc->sc_aica_regh,
		    AICA_LOOP_END_REG(ch), 0);

		bus_space_write_4(sc->sc_memt,sc->sc_aica_regh,
		    AICA_ENVELOPE1_REG(ch), 0x1f);
		bus_space_write_4(sc->sc_memt,sc->sc_aica_regh,
		    AICA_ENVELOPE2_REG(ch), 0x1f);

		bus_space_write_4(sc->sc_memt,sc->sc_aica_regh,
		    AICA_LFO1_REG(ch), 0);
		bus_space_write_4(sc->sc_memt,sc->sc_aica_regh,
		    AICA_DSPCFG_REG(ch), 0);

		bus_space_write_4(sc->sc_memt,sc->sc_aica_regh,
		    AICA_VOLPAN_REG(ch), 0);

		bus_space_write_4(sc->sc_memt, sc->sc_aica_regh,
		    AICA_TL_Q_REG(ch), 32);

		bus_space_write_4(sc->sc_memt,sc->sc_aica_regh,
		    AICA_FLV0_REG(ch), 0);
		bus_space_write_4(sc->sc_memt,sc->sc_aica_regh,
		    AICA_FLV1_REG(ch), 0);
		bus_space_write_4(sc->sc_memt,sc->sc_aica_regh,
		    AICA_FLV2_REG(ch), 0);
		bus_space_write_4(sc->sc_memt,sc->sc_aica_regh,
		    AICA_FLV3_REG(ch), 0);
		bus_space_write_4(sc->sc_memt,sc->sc_aica_regh,
		    AICA_FLV4_REG(ch), 0);
		bus_space_write_4(sc->sc_memt,sc->sc_aica_regh,
		    AICA_AR_D1R_REG(ch), 0);
		bus_space_write_4(sc->sc_memt,sc->sc_aica_regh,
		    AICA_D2R_RR_REG(ch), 0);
	}

	/* init variables */
	sc->sc_msgpos = 0;

	/* init mastervolume */
	sc->sc_output_master = 255;
	aica_set_mvol(sc, sc->sc_output_master);

	/* init channel volume */
	for (n = 0; n < AICA_NAUDIO; n++) {
		sc->sc_audioarg[n].aaa_gain_left = 255;
		sc->sc_audioarg[n].aaa_gain_right = 255;
		aica_set_vol(sc, n,
		    sc->sc_audioarg[n].aaa_gain_left,
		    sc->sc_audioarg[n].aaa_gain_right);
	}

}

void
aica_load_microcode(struct aica_softc *sc, void *code, uint32_t codesize)
{
	/*
	 * clear AICA memory and load microcode
	 */
	bus_space_set_region_4(sc->sc_memt, sc->sc_aica_memh,
	    0, 0, AICA_RAM_SIZE / 4);
	aica_memwrite(sc, AICA_ARM_CODE, code, codesize);
}

void
aica_enable(struct aica_softc *sc)
{
	uint32_t v;
	int i;

	if (sc->sc_aica_enabled != 0)
		return;

	log(LOG_DEBUG, "INIT aica arm\n");

	sc->sc_aica_enabled = 1;

	/* initialize SH4-ARM command queue */
	aica_initqueue(sc);

	/* set CPU clock 25MHz */
	bus_space_write_4(sc->sc_memt, sc->sc_aica_regh, AICA_RSVD1_REG, 0x80);
	bus_space_write_4(sc->sc_memt, sc->sc_aica_regh, AICA_CLOCK_REG, 24);

	/* clear reset */
	v = bus_space_read_4(sc->sc_memt, sc->sc_aica_regh, AICA_RESETCTL_REG);
	bus_space_write_4(sc->sc_memt, sc->sc_aica_regh, AICA_RESETCTL_REG,
	    v & ~AICA_RESETCTL_RESET);

	/* confirm heartbeat */
	v = bus_space_read_4(sc->sc_memt, sc->sc_aica_memh, AICA_HEARTBEAT);
	for (i = 0; i < 200; i++) {
		if (v != bus_space_read_4(sc->sc_memt, sc->sc_aica_memh, AICA_HEARTBEAT))
			break;
		delay(100);
	}
	log(LOG_DEBUG, "aica boot ready delay: %d\n", i);
}

void
aica_disable(struct aica_softc *sc)
{
	uint32_t v;

	log(LOG_DEBUG, "disable aica arm\n");

	/* hold reset */
	v = bus_space_read_4(sc->sc_memt, sc->sc_aica_regh, AICA_RESETCTL_REG);
	bus_space_write_4(sc->sc_memt, sc->sc_aica_regh, AICA_RESETCTL_REG,
	    v | AICA_RESETCTL_RESET);

	sc->sc_aica_enabled = 0;
}

void
aica_memread(struct aica_softc *sc, bus_size_t offset, uint32_t *dst, int len)
{
	KASSERT((offset & 3) == 0);
	KASSERT((len & 3) == 0);

	bus_space_read_region_4(sc->sc_memt, sc->sc_aica_memh,
	    offset, dst, (len + 3) / 4);
}

void
aica_memwrite(struct aica_softc *sc, bus_size_t offset, const uint32_t *src, int len)
{
	KASSERT((offset & 3) == 0);

	bus_space_write_region_4(sc->sc_memt, sc->sc_aica_memh,
	    offset, src, (len + 3) / 4);
}

void
aica_set_mvol(struct aica_softc *sc, int vol)
{
	bus_space_write_4(sc->sc_memt, sc->sc_aica_regh,
	    AICA_VOLUME_REG, LEVEL256TO16(vol));
}

void
aica_set_vol(struct aica_softc *sc, int unit, int left, int right)
{
	if (sc->sc_audioarg[unit].aaa_channels > 1) {
		bus_space_write_4(sc->sc_memt, sc->sc_aica_regh,
		    AICA_VOLPAN_REG(unit * 2 + 0),
		    AICA_VOLPAN_VOL(LEVEL256TO16(left)) |
		    AICA_VOLPAN_PAN(0x1f));	/* left pan */

		bus_space_write_4(sc->sc_memt, sc->sc_aica_regh,
		    AICA_VOLPAN_REG(unit * 2 + 1),
		    AICA_VOLPAN_VOL(LEVEL256TO16(right)) |
		    AICA_VOLPAN_PAN(0x0f));	/* right pan */
	} else {
		bus_space_write_4(sc->sc_memt, sc->sc_aica_regh,
		    AICA_VOLPAN_REG(unit * 2 + 0),
		    AICA_VOLPAN_VOL(LEVEL256TO16(left)) |
		    AICA_VOLPAN_PAN(0));	/* middle pan */
	}
}

/*
 * read msgbuf in AICA memory
 */
void
aica_msgbuf(struct aica_softc *sc)
{
	int pos;
	int append;
	char ch;

	aica_memread(sc, AICA_MSGBUF, (uint32_t *)sc->sc_msgbuf, AICA_MSGBUFSIZE);
	sc->sc_msgbuf[AICA_MSGBUFSIZE] = '\0';

	append = 0;
	pos = sc->sc_msgpos;
	for (;;) {
		ch = sc->sc_msgbuf[pos];

		if (ch == '\0') {
			if (sc->sc_msgpos == pos)
				break;

			if (append) {
				addlog("%s", &sc->sc_msgbuf[sc->sc_msgpos]);
				append = 0;
			} else {
				log(LOG_DEBUG, "%s", &sc->sc_msgbuf[sc->sc_msgpos]);
			}
			sc->sc_msgpos = pos;
			break;
		}

		if (ch == '\n') {
			sc->sc_msgbuf[pos++] = '\0';
			if (append) {
				addlog("%s\n", &sc->sc_msgbuf[sc->sc_msgpos]);
				append = 0;
			} else {
				log(LOG_DEBUG, "%s\n", &sc->sc_msgbuf[sc->sc_msgpos]);
			}

			if (pos >= AICA_MSGBUFSIZE) {
				pos = 0;
			}
			sc->sc_msgpos = pos;
			continue;
		}

		pos++;
		if (pos >= AICA_MSGBUFSIZE) {
			log(LOG_DEBUG, "%s", &sc->sc_msgbuf[sc->sc_msgpos]);
			append = 1;
			sc->sc_msgpos = 0;
			continue;
		}
	}
}

struct cmd *
aica_readcmd(struct aica_softc *sc, struct cmdqueue *cmdqueue)
{
	uint32_t cmdoff, serial;
	int s;

	s = splaudio();

	cmdoff = cmdqueue->address + cmdqueue->index * sizeof(struct cmd);
//	log(LOG_DEBUG, "cmdoff=%d, cmdbuf=%p, cmdsize=%d\n", cmdoff,
//	    &sc->sc_cmdbuf, sizeof(struct cmd));

	aica_memread(sc, cmdoff, (uint32_t *)&sc->sc_cmdbuf,
	    sizeof(struct cmd));

//	xxxdump(&sc->sc_cmdbuf, sizeof(struct cmd));

	/* don't read incomplete message */
	if (sc->sc_cmdbuf.magic != CMDMAGIC) {
//		log(LOG_DEBUG, "no magic. ignore\n");
		splx(s);
		return NULL;
	}

	/* don't read forepast message */
	serial = sc->sc_cmdbuf.serial;
	serial -= cmdqueue->serial;
	if (serial >= 0x80000000) {
		splx(s);
		return NULL;
	}

	/* clear magic in queue */
	bus_space_write_4(sc->sc_memt, sc->sc_aica_memh,
	    cmdoff + offsetof(struct cmd, magic), 0);

#if 0
	{
		/* verify */
		uint32_t v;
		v = bus_space_read_4(sc->sc_memt, sc->sc_aica_memh,
		    cmdoff + offsetof(struct cmd, magic));
		if (v != 0) {
			log(LOG_DEBUG,
			    "magic cleared, but magic not cleared: 0x%08x\n", v);
		}
	}
#endif

	/* next serial to read */
	cmdqueue->serial = sc->sc_cmdbuf.serial + 1;

	/* forward index */
	cmdqueue->index++;
	if ((cmdqueue->index * sizeof(struct cmd)) >= cmdqueue->size)
		cmdqueue->index = 0;

	splx(s);

	return &sc->sc_cmdbuf;
}

void
aica_sendcmd(struct aica_softc *sc, struct cmdqueue *cmdqueue, struct cmd *cmd)
{
	uint32_t cmdoff;
	struct cmd tmpcmd;
	int s;

	tmpcmd.command = cmd->command;
	tmpcmd.param = cmd->param;
	tmpcmd.serial = cmdqueue->serial;
	tmpcmd.magic = CMDMAGIC;

	s = splaudio();

	/* write to AICA */
	cmdoff = cmdqueue->address + cmdqueue->index * sizeof(struct cmd);
	bus_space_write_region_4(sc->sc_memt, sc->sc_aica_memh,
	    cmdoff, (uint32_t *)&tmpcmd, sizeof(struct cmd) / 4);

	/* next serial to write */
	cmdqueue->serial++;

	/* forward index */
	cmdqueue->index++;
	if ((cmdqueue->index * sizeof(struct cmd)) >= cmdqueue->size)
		cmdqueue->index = 0;

	splx(s);
}

void
aica_sendcmd_channel_rate(struct aica_softc *sc, int unit, int ch, int rate)
{
	struct cmd tmpcmd;

	tmpcmd.command = AICA_COMMAND_PACK(AICA_CMD_CH_RATE, unit);
	tmpcmd.param = AICA_PARAM_PACK(ch, rate);
	aica_sendcmd(sc, &sc->sc_cmdq_to_aica, &tmpcmd);
}

void
aica_sendcmd_prec_blksize(struct aica_softc *sc, int unit,
                          int precision, int blksize_per_ch)
{
	struct cmd tmpcmd;

	tmpcmd.command = AICA_COMMAND_PACK(AICA_CMD_PREC_BLKSIZE, unit);
	tmpcmd.param = AICA_PARAM_PACK(precision, blksize_per_ch);
	aica_sendcmd(sc, &sc->sc_cmdq_to_aica, &tmpcmd);
}

void
aica_sendcmd_nblk_nsample(struct aica_softc *sc, int unit,
                          int nblk, int nsample)
{
	struct cmd tmpcmd;

	tmpcmd.command = AICA_COMMAND_PACK(AICA_CMD_NBLK_NSMPL, unit);
	tmpcmd.param = AICA_PARAM_PACK(nblk, nsample);
	aica_sendcmd(sc, &sc->sc_cmdq_to_aica, &tmpcmd);
}

void
aica_sendcmd_play(struct aica_softc *sc, int unit)
{
	struct cmd tmpcmd;

	tmpcmd.command = AICA_COMMAND_PACK(AICA_CMD_PLAY, unit);
	tmpcmd.param = 0;
	aica_sendcmd(sc, &sc->sc_cmdq_to_aica, &tmpcmd);
}

void
aica_sendcmd_stop(struct aica_softc *sc, int unit)
{
	struct cmd tmpcmd;

	tmpcmd.command = AICA_COMMAND_PACK(AICA_CMD_STOP, unit);
	tmpcmd.param = 0;
	aica_sendcmd(sc, &sc->sc_cmdq_to_aica, &tmpcmd);

#if 0
	bus_space_write_4(sc->sc_memt, sc->sc_aica_regh,
	    AICA_PLAYCTRL_REG(unit * 2 + 0), 0);
	bus_space_write_4(sc->sc_memt, sc->sc_aica_regh,
	    AICA_PLAYCTRL_REG(unit * 2 + 1), 0);
#endif
}

/*
 * G2 bus cycles must not be interrupted by IRQs or G2 DMA.
 * The following paired macros will take the necessary precautions.
 */
#define G2LOCK_DECL							\
	int __s

#define G2_LOCK()							\
	do {								\
		__s = _cpu_intr_suspend();				\
		/* suspend any G2 DMA here... */			\
		while ((*(volatile uint32_t *)0xa05f688c) & 0x20)	\
			;						\
	} while (/*CONSTCOND*/0)

#define G2_UNLOCK()							\
	do {								\
		/* resume any G2 DMA here... */				\
		_cpu_intr_resume(__s);					\
	} while (/*CONSTCOND*/0)


/*
 * split 2channel PCM stream to L and R buffers
 *
 * We should do split channel stream to bounce buffer once,
 * and write to AICA memory with bus_space_write_region_4().
 *
 * XXX:
 *   But it is too slow and additional memory copy.
 *   split channel and write to AICA memory **DIRECTLY**
 */
void
aica_ch2p16write(struct aica_softc *sc, bus_size_t offset, uint16_t *src,
    int len)
{
	volatile uint32_t *dst;
	uint32_t data;
	G2LOCK_DECL;

	KASSERT((offset & 3) == 0);

	/* XXX: bus_space_handler violation, but fast */
	dst = (uint32_t *)(sc->sc_aica_memh + offset);

	G2_LOCK();

	for (; len >= 32; len -= 32) {
#define	CH2P16WRITE4			\
		data = *src++ << 16;	\
		src++;			\
		data |= *src++;		\
		src++;			\
		*dst++ = data;

		CH2P16WRITE4
		CH2P16WRITE4
		CH2P16WRITE4
		CH2P16WRITE4

		CH2P16WRITE4
		CH2P16WRITE4
		CH2P16WRITE4
		CH2P16WRITE4
	}

	for (; len >= 4; len -= 4) {
		CH2P16WRITE4
	}

	if (len >= 2)
		*(volatile uint16_t *)dst++ = *src++;

	G2_UNLOCK();
}

void
aica_ch2p8write(struct aica_softc *sc, bus_size_t offset, uint8_t *src,
    int len)
{
	G2LOCK_DECL;
	volatile uint32_t *dst;
	uint32_t data;

	KASSERT((offset & 3) == 0);

	/* XXX: bus_space_handler violation, but fast */
	dst = (uint32_t *)(sc->sc_aica_memh + offset);

	G2_LOCK();

	for (; len >= 32; len -= 32) {
#define	CH2P8WRITE4			\
		data = *src++ << 24;	\
		src++;			\
		data |= *src++ << 16;	\
		src++;			\
		data |= *src++ << 8;	\
		src++;			\
		data |= *src++;		\
		src++;			\
		*dst++ = data;

		CH2P8WRITE4
		CH2P8WRITE4
		CH2P8WRITE4
		CH2P8WRITE4

		CH2P8WRITE4
		CH2P8WRITE4
		CH2P8WRITE4
		CH2P8WRITE4
	}

	for (; len >= 4; len -= 4) {
		CH2P8WRITE4
	}

	for (; len > 0; len--) {
		*(volatile uint8_t *)dst++ = *src++;
		src++;
	}

	G2_UNLOCK();
}

/* ARGSUSED */
int
aica_open(void *addr, int flags __unused)
{
	struct aica_audioarg *aaa;
	struct aica_softc *sc;

	aaa = addr;
	sc = aaa->aaa_sc;

	log(LOG_DEBUG, "%s:%d: unit=%d\n", __func__, __LINE__, aaa->aaa_unit);

	if (aaa->aaa_open)
		return EBUSY;

	aaa->aaa_intr = NULL;
	aaa->aaa_open = 1;
	aaa->aaa_playing = 0;

	aica_enable(sc);	/* Run AICA ARM7 CPU */

	aica_heartbeat(sc);

	return 0;
}

void
aica_close(void *addr)
{
	struct aica_audioarg *aaa;
	struct aica_softc *sc;

	aaa = addr;
	sc = aaa->aaa_sc;

	log(LOG_DEBUG, "%s:%d: unit=%d\n", __func__, __LINE__, aaa->aaa_unit);

	aica_heartbeat(sc);

	aaa->aaa_intr = NULL;
	aaa->aaa_open = 0;
	aaa->aaa_playing = 0;
}

/* ARGSUSED */
int
aica_query_encoding(void *addr, struct audio_encoding *fp)
{
	struct aica_audioarg *aaa;
	struct aica_softc *sc;

	aaa = addr;
	sc = aaa->aaa_sc;

	return auconv_query_encoding(sc->sc_encodings, fp);
}

/* ARGSUSED */
int
aica_set_params(void *addr, int setmode __unused, int usemode __unused,
                audio_params_t *play, audio_params_t *rec __unused,
                stream_filter_list_t *pfil, stream_filter_list_t *rfil __unused)
{
	struct aica_audioarg *aaa;
	struct aica_softc *sc;
	const audio_params_t *hw;
	int i;

	aaa = addr;
	sc = aaa->aaa_sc;

	log(LOG_DEBUG, "%s:%d: unit=%d\n", __func__, __LINE__, aaa->aaa_unit);

	i = auconv_set_converter(aica_formats, __arraycount(aica_formats),
	    AUMODE_PLAY, play, TRUE, pfil);
	if (i < 0)
		return EINVAL;

	hw = pfil->req_size > 0 ? &pfil->filters[0].param : play;

	aaa->aaa_precision = hw->precision;
	aaa->aaa_channels = hw->channels;
	aaa->aaa_rate = hw->sample_rate;

	log(LOG_DEBUG, "%s:%d: ch=%d prec=%d, rate=%d\n", __func__, __LINE__,
	    aaa->aaa_channels,
	    aaa->aaa_precision,
	    aaa->aaa_rate);

	return 0;
}

int
aica_blocksize(struct aica_softc *sc, int unit)
{
	int size;

	log(LOG_DEBUG, "%s:%d: unit=%d\n", __func__, __LINE__, unit);

	switch (sc->sc_audioarg[unit].aaa_precision) {
	case 4:
		if (sc->sc_audioarg[unit].aaa_channels == 1)
			size = AICA_DMABUF_SIZE / 4;
		else
			size = AICA_DMABUF_SIZE / 2;
		break;
	case 8:
		if (sc->sc_audioarg[unit].aaa_channels == 1)
			size = AICA_DMABUF_SIZE / 2;
		else
			size = AICA_DMABUF_SIZE;
		break;
	case 16:
		if (sc->sc_audioarg[unit].aaa_channels == 1)
			size = AICA_DMABUF_SIZE;
		else
			size = AICA_DMABUF_SIZE * 2;
		break;
	default:
		size = AICA_DMABUF_SIZE / 4;
		break;
	}

	return size;
}

/* ARGSUSED */
int
aica_round_blocksize(void *addr, int blk, int mode __unused,
                     const audio_params_t *param __unused)
{
	struct aica_audioarg *aaa;
	struct aica_softc *sc;
	int size;

	aaa = addr;
	sc = aaa->aaa_sc;

	log(LOG_DEBUG, "%s:%d: unit=%d\n", __func__, __LINE__, aaa->aaa_unit);

	size = aica_blocksize(sc, aaa->aaa_unit);
	if (size > blk)
		size = blk;

	log(LOG_DEBUG, "%s:%d: blk=%d -> %d\n", __func__, __LINE__, blk, size);

	return size;
}

void
aica_blockdone(struct aica_audioarg *aaa)
{
	if (aaa->aaa_usedblk > 0)
		aaa->aaa_usedblk--;
}

void
aica_fillbuffer(struct aica_softc *sc, struct aica_audioarg *aaa, void *buffer)
{
//	log(LOG_DEBUG, "%s:%d: unit=%d\n", __func__, __LINE__, aaa->aaa_unit);

	switch (aaa->aaa_channels) {
	case 1:
		aica_memwrite(sc,
		    AICA_DMABUF(aaa->aaa_unit) +
		    aaa->aaa_nextblk * aaa->aaa_blksize_per_ch,
		    buffer, aaa->aaa_blksize_per_ch);
		break;
	case 2:
		/* split stereo stream to each L and R buffers */
		switch (aaa->aaa_precision) {
		case 8:
			aica_ch2p8write(sc,
			    AICA_DMABUF_LEFT(aaa->aaa_unit) +
			    aaa->aaa_nextblk * aaa->aaa_blksize_per_ch,
			    (uint8_t *)buffer + 0, aaa->aaa_blksize_per_ch);
			aica_ch2p8write(sc,
			    AICA_DMABUF_RIGHT(aaa->aaa_unit) +
			    aaa->aaa_nextblk * aaa->aaa_blksize_per_ch,
			    (uint8_t *)buffer + 1, aaa->aaa_blksize_per_ch);
			break;
		case 16:
			aica_ch2p16write(sc,
			    AICA_DMABUF_LEFT(aaa->aaa_unit) +
			    aaa->aaa_nextblk * aaa->aaa_blksize_per_ch,
			    (uint16_t *)buffer + 0, aaa->aaa_blksize_per_ch);
			aica_ch2p16write(sc,
			    AICA_DMABUF_RIGHT(aaa->aaa_unit) +
			    aaa->aaa_nextblk * aaa->aaa_blksize_per_ch,
			    (uint16_t *)buffer + 1, aaa->aaa_blksize_per_ch);
			break;
		}
		break;
	}

	aaa->aaa_usedblk++;

	log(LOG_DEBUG, "%s:%d: unit=%d, ch=%d, prec=%d, rate=%d, blksize_per_ch=%d, blk=%d/%d nsamples=%d\n",
	    __func__, __LINE__,
	    aaa->aaa_unit, aaa->aaa_channels, aaa->aaa_precision, aaa->aaa_rate, aaa->aaa_blksize_per_ch,
	    aaa->aaa_usedblk, aaa->aaa_nblk, aaa->aaa_nsamples);

	/* forward block */
	if (++aaa->aaa_nextblk >= aaa->aaa_nblk)
		aaa->aaa_nextblk = 0;
}

void
aica_proceed(struct aica_softc *sc)
{
	struct aica_audioarg *aaa;
	struct cmd *cmd;
	int unit, i, nfwd;

//	aica_heartbeat(sc);

	while ((cmd = aica_readcmd(sc, &sc->sc_cmdq_from_aica)) != NULL) {
//		log(LOG_DEBUG, "command=%d, param=%d, serial=0x%04x, magic=0x%04x\n",
//		    cmd->command,
//		    cmd->param,
//		    cmd->serial,
//		    cmd->magic);

		switch (AICA_COMMAND_CMD(cmd->command)) {
		case AICA_CMD_MSGBUF:
//			log(LOG_DEBUG, "MSGBUF: serial=%d param=%d\n",
//			    cmd->serial, cmd->param);
			aica_msgbuf(sc);
			break;
		case AICA_CMD_FORWARD:
			unit = AICA_COMMAND_UNIT(cmd->command);
			if (unit >= AICA_NAUDIO) {	/* illegal parameter */
				log(LOG_DEBUG, "illegal unit: AICA_COMMAND_FORWARD: unit=%d\n", unit);
				break;
			}

			aaa = &sc->sc_audioarg[unit];
			nfwd = cmd->param;
			if (nfwd > aaa->aaa_nblk) {
				log(LOG_DEBUG, "illegal parameter: AICA_COMMAND_FORWARD: %d\n", nfwd);
				break;
			}

			for (i = 0; i < nfwd; i++) {
				aica_blockdone(aaa);

				/* call audio callback handler; audio_pint() */
				if (aaa->aaa_open && aaa->aaa_intr != NULL)
					(*(aaa->aaa_intr))(aaa->aaa_intr_arg);
			}
			break;
		default:
			break;
		}
	}
}

int
aica_intr(void *arg)
{
	struct aica_softc *sc;

	sc = arg;

//	log(LOG_NOTICE, "%s:%d\n", __func__, __LINE__);
	aica_proceed(sc);

	/* clear AICA interrupt */
	bus_space_write_4(sc->sc_memt, sc->sc_aica_regh,
	    AICA_INTRST_REG, AICA_INT_CPU);

	return 1;
}

int
aica_init_output(void *addr, void *buffer, int size)
{
	struct aica_audioarg *aaa;
	struct aica_softc *sc;

	aaa = addr;
	sc = aaa->aaa_sc;

	log(LOG_DEBUG, "%s:%d: unit=%d, buffer=%p, size=%d\n", __func__, __LINE__,
	    aaa->aaa_unit, buffer, size);

	sc->sc_buffersize = size;
	aaa->aaa_nextblk = 0;
	aaa->aaa_usedblk = 0;

	aica_heartbeat(sc);

	return 0;
}

int
aica_start_output(void *addr, void *block, int blksize,
                  void (*intr)(void*), void *intrarg)
{
	struct aica_audioarg *aaa;
	struct aica_softc *sc;

	aaa = addr;
	sc = aaa->aaa_sc;

	log(LOG_DEBUG, "%s:%d: unit=%d, block=%p, blksize=%d\n", __func__, __LINE__,
	    aaa->aaa_unit, block, blksize);

//	aica_heartbeat(sc);

	aaa->aaa_intr = intr;
	aaa->aaa_intr_arg = intrarg;

	if (aaa->aaa_usedblk == 0) {
		aaa->aaa_blksize_per_ch = blksize / aaa->aaa_channels;
		aaa->aaa_nsamples = aaa->aaa_blksize_per_ch * 8 / aaa->aaa_precision;

		aaa->aaa_nblk = AICA_DMABUF_SIZE / aaa->aaa_nsamples;
		if (aaa->aaa_nblk > (sc->sc_buffersize / blksize))
			aaa->aaa_nblk = sc->sc_buffersize / blksize;

		if (aaa->aaa_nblk > 4)
			aaa->aaa_nblk = 4;

		aica_sendcmd_channel_rate(sc, aaa->aaa_unit,
		    aaa->aaa_channels, aaa->aaa_rate);
		aica_sendcmd_prec_blksize(sc, aaa->aaa_unit,
		    aaa->aaa_precision, aaa->aaa_blksize_per_ch);
		aica_sendcmd_nblk_nsample(sc, aaa->aaa_unit,
		    aaa->aaa_nblk, aaa->aaa_nsamples);
	}

	aica_fillbuffer(sc, aaa, block);

	if (aaa->aaa_nblk > aaa->aaa_usedblk) {
		/* get more data */
		if (aaa->aaa_intr != NULL)
			(*(aaa->aaa_intr))(aaa->aaa_intr_arg);
	}

	if (aaa->aaa_playing == 0) {
		aica_sendcmd_play(sc, aaa->aaa_unit);
		aaa->aaa_playing = 1;
	}

	return 0;
}

/* ARGSUSED */
int
aica_init_input(void *addr __unused, void *buffer __unused, int size __unused)
{
	return ENODEV;
}

/* ARGSUSED */
int
aica_start_input(void *addr __unused, void *blk __unused, int blksize __unused,
                 void (*intr)(void*) __unused, void *intrarg __unused)
{
	return ENODEV;
}

int
aica_halt_output(void *addr)
{
	struct aica_audioarg *aaa;
	struct aica_softc *sc;

	aaa = addr;
	sc = aaa->aaa_sc;

	aica_heartbeat(sc);

	log(LOG_DEBUG, "%s:%d: unit=%d, playing=%d\n", __func__, __LINE__,
	    aaa->aaa_unit, aaa->aaa_playing);

	if (aaa->aaa_playing != 0) {
		aica_sendcmd_stop(sc, aaa->aaa_unit);
		aaa->aaa_playing = 0;
	}

	return 0;
}

/* ARGSUSED */
int
aica_halt_input(void *addr __unused)
{
	return ENODEV;
}

/* ARGSUSED */
int
aica_getdev(void *addr __unused, struct audio_device *ret)
{
	*ret = aica_device;
	return 0;
}

int
aica_set_port(void *addr, mixer_ctrl_t *mc)
{
	struct aica_audioarg *aaa;
	struct aica_softc *sc;

	aaa = addr;
	sc = aaa->aaa_sc;

	switch (mc->dev) {
	case AICA_MASTER_VOL:
		sc->sc_output_master =
		    mc->un.value.level[AUDIO_MIXER_LEVEL_MONO] & 0xff;

		aica_set_mvol(sc, sc->sc_output_master);
		break;
	case AICA_OUTPUT_GAIN:
		aaa->aaa_gain_left =
		    mc->un.value.level[AUDIO_MIXER_LEVEL_LEFT] & 0xff;
		aaa->aaa_gain_right =
		    mc->un.value.level[AUDIO_MIXER_LEVEL_RIGHT] & 0xff;

		aica_set_vol(sc, aaa->aaa_unit,
		    aaa->aaa_gain_left, aaa->aaa_gain_right);
		break;
	default:
		return EINVAL;
	}

	return 0;
}

int
aica_get_port(void *addr, mixer_ctrl_t *mc)
{
	struct aica_audioarg *aaa;
	struct aica_softc *sc;

	aaa = addr;
	sc = aaa->aaa_sc;

	switch (mc->dev) {
	case AICA_MASTER_VOL:
		if (mc->un.value.num_channels != 1)
			return EINVAL;
		mc->un.value.level[AUDIO_MIXER_LEVEL_MONO] =
		    LEVEL16TO256(LEVEL256TO16(sc->sc_output_master));
		break;
	case AICA_OUTPUT_GAIN:
		mc->un.value.level[AUDIO_MIXER_LEVEL_LEFT] =
		    aaa->aaa_gain_left;
		mc->un.value.level[AUDIO_MIXER_LEVEL_RIGHT] =
		    aaa->aaa_gain_right;
		break;
	default:
		return EINVAL;
	}
	return 0;
}

/* ARGSUSED */
int
aica_query_devinfo(void *addr __unused, mixer_devinfo_t *md)
{
	struct aica_audioarg *aaa;
	struct aica_softc *sc;

	aaa = addr;
	sc = aaa->aaa_sc;

	switch (md->index) {
	case AICA_MASTER_VOL:
		md->type = AUDIO_MIXER_VALUE;
		md->mixer_class = AICA_OUTPUT_CLASS;
		md->prev = md->next = AUDIO_MIXER_LAST;
		strcpy(md->label.name, AudioNmaster);
		md->un.v.num_channels = 1;
		strcpy(md->un.v.units.name, AudioNvolume);
		return 0;
	case AICA_OUTPUT_GAIN:
		md->type = AUDIO_MIXER_VALUE;
		md->mixer_class = AICA_OUTPUT_CLASS;
		md->prev = md->next = AUDIO_MIXER_LAST;
		strcpy(md->label.name, AudioNoutput);
		md->un.v.num_channels = 2;
		strcpy(md->label.name, AudioNvolume);
		return 0;
	case AICA_OUTPUT_CLASS:
		md->type = AUDIO_MIXER_CLASS;
		md->mixer_class = AICA_OUTPUT_CLASS;
		md->next = md->prev = AUDIO_MIXER_LAST;
		strcpy(md->label.name, AudioCoutputs);
		return 0;
	}

	return ENXIO;
}

/* ARGSUSED */
int
aica_get_props(void *addr __unused)
{
	return AUDIO_PROP_PLAYBACK;
}
