/*	$NetBSD: aica_arm.c,v 1.4 2005/12/24 20:06:59 perry Exp $	*/

/*
 * Copyright (c) 2003 SHIMIZU Ryo <ryo@misakimix.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "aica_arm.h"
#include <sys/null.h>
#include <arch/dreamcast/dev/g2/aicavar.h>
#include <arch/dreamcast/dev/g2/aicareg.h>

#undef AICA_DEBUG
#define AICA_PRINTF

#define REGBASE	AICA_REG_ADDR_ARM
#define	REG_READ_1(off)		(*(volatile uint8_t  *)(REGBASE + (off)))
#define	REG_READ_2(off)		(*(volatile uint16_t *)(REGBASE + (off)))
#define	REG_READ_4(off)		(*(volatile uint32_t *)(REGBASE + (off)))
#define	REG_WRITE_1(off, v)	*(volatile uint8_t  *)(REGBASE + (off)) = (v)
#define	REG_WRITE_2(off, v)	*(volatile uint16_t *)(REGBASE + (off)) = (v)
#define	REG_WRITE_4(off, v)	*(volatile uint32_t *)(REGBASE + (off)) = (v)
#define	CH_READ_4(ch, off)	REG_READ_4(((ch) << 7) + (off))
#define	CH_WRITE_4(ch, off, v)	REG_WRITE_4(((ch) << 7) + (off), v)

void aica_init(void);
void aica_play(int);
void aica_stop(int);
void aica_main(void);

static void bzero32(void *, int);

uint32_t rate2reg(unsigned int);
struct cmd *aica_readcmd(struct cmdqueue *);
void aica_sendcmd(struct cmdqueue *, struct cmd *);

#define MAXUNIT	4

struct unitwork {
	uint32_t state;
#define UNIT_STATE_IDLE		0
#define UNIT_STATE_PLAYING	1
#define UNIT_STATE_STOPPING	2
	uint32_t channel;
	uint32_t rate;
	uint32_t precision;
	uint32_t curpos;
	uint32_t nsample;	/* nsample per block */
	uint32_t nblk;
	uint32_t blksize;	/* bytes per block */
	int32_t leftblk;	/* left to stop */
	uint32_t leftvol;
	uint32_t rightvol;
} unitwork[MAXUNIT];


struct cmd *
aica_readcmd(struct cmdqueue *cmdqueue)
{
	struct cmd *cmd;
	uint32_t serial;

	cmd = (struct cmd *)(cmdqueue->address +
	    (cmdqueue->index * sizeof(struct cmd)));

	/* dont read incomplete message */
	if (cmd->magic != CMDMAGIC)
		return NULL;

	/* dont read forepast message */
	serial = cmd->serial;
	serial -= cmdqueue->serial;
	if (serial >= 0x8000000)
		return NULL;

	cmd->magic = 0;

	/* next serial to read */
	cmdqueue->serial = cmd->serial + 1;

	/* forward index */
	cmdqueue->index++;
	if ((cmdqueue->index * sizeof(struct cmd)) >= cmdqueue->size)
		cmdqueue->index = 0;

	return cmd;
}

void
aica_sendcmd(struct cmdqueue *cmdqueue, struct cmd *cmd)
{
	struct cmd *cmdp;

	cmdp = (struct cmd *)(cmdqueue->address +
	    (cmdqueue->index * sizeof(struct cmd)));

	cmdp->command = cmd->command;
	cmdp->param = cmd->param;
	cmdp->serial = cmdqueue->serial;
	cmdp->magic = CMDMAGIC;

	/* next serial to write */
	cmdqueue->serial++;

	/* forward index */
	cmdqueue->index++;
	if ((cmdqueue->index * sizeof(struct cmd)) >= cmdqueue->size)
		cmdqueue->index = 0;
}

void
aica_init(void)
{
	int ch, off;

	/* Initialize AICA channels */
//	REG_WRITE_4(AICA_VOLUME_REG, 0x0000);	/* Master volume: Min */

	for (ch = 0; ch < 64; ch++) {
		CH_WRITE_4(ch, 0x00, 0x8000);	/* Key off */
		CH_WRITE_4(ch, 0x04, 0x0000);	/* DataAddress (low) */
		CH_WRITE_4(ch, 0x08, 0x0000);	/* LoopStartPosition */
		CH_WRITE_4(ch, 0x0c, 0x0000);	/* LoopEndPosition */
		CH_WRITE_4(ch, 0x10, 0x001f);	/* AR = 0x1f = 0 msec */
		CH_WRITE_4(ch, 0x14, 0x001f);	/* RR = 0x1f = 0 msec */
		CH_WRITE_4(ch, 0x18, 0x0000);	/* Pitch */
		CH_WRITE_4(ch, 0x1c, 0x0000);	/* LFO Control */
		CH_WRITE_4(ch, 0x20, 0x0000);	/* DSP Channel to send */
		CH_WRITE_4(ch, 0x24, 0x0000);	/* Pan & Volume */
//		CH_WRITE_4(ch, 0x28, 0x0024);	/* Volume & LowPassFilter */
		CH_WRITE_4(ch, 0x2c, 0x0000);	/* LowPassFilter for Attack  */
		CH_WRITE_4(ch, 0x30, 0x0000);	/* LowPassFilter for Decay   */
		CH_WRITE_4(ch, 0x34, 0x0000);	/* LowPassFilter for Sustain */
		CH_WRITE_4(ch, 0x38, 0x0000);	/* LowPassFilter for Keyoff  */
		CH_WRITE_4(ch, 0x3c, 0x0000);	/* LowPassFilter for Release */
		CH_WRITE_4(ch, 0x40, 0x0000);	/* LowPassFilter transition
						   for Attack, Decay */
		CH_WRITE_4(ch, 0x44, 0x0000);	/* LowPassFilter transition
						   for Decay, Release */

		for (off = 0x48; off < 0x80; off+=4) {
			CH_WRITE_4(ch, off, 0x0000);	/* other = 0 */
		}
	}
//	REG_WRITE_4(AICA_VOLUME_REG, 0x000f);	/* Master volume: Max */
}


static void
bzero32(void *b, int len)
{
	for (; len > 0; len -= 4)
		*(uint32_t *)b++ = 0;
}

uint32_t
rate2reg(unsigned int rate)
{
	uint32_t base, fns;
	int oct;

	base = 44100 << 7;
	for (oct = 7; oct >= -8 && rate < base; oct--)
		base >>= 1;

	if (rate < base)
		return (oct << 11) & 0xf800;

	rate -= base;

#if 1
	/* (base / 2) : round off */
	fns = (rate * 1024 + (base / 2)) / base;
#else
	/* avoid using udivsi3() */
	{
		uint32_t tmp = (rate * 1024 + (base / 2));
		for (fns = 0; tmp >= base; tmp -= base, fns++)
			;
	}
#endif

	/* adjustment */
	if (fns == 1024) {
		oct++;
		fns = 0;
	} else {
		if ((rate > base * fns / 1024) &&
		    (fns < 1023) &&
		    (rate == base * (fns + 1) / 1024)) {
			fns++;
		} else if ((rate < base * fns / 1024) &&
		    (fns > 0) &&
		    (rate == base * (fns - 1)/ 1024)) {
			fns--;
		}
	}

	return ((oct << 11) & 0xf800) + fns;
}

void
aica_play(int unit)
{
	int lch, rch;
	uint32_t mode, ratepitch;

	lch = unit * 2 + 0;
	rch = unit * 2 + 1;

	REG_WRITE_4(AICA_PLAYCTRL_REG(lch), AICA_PLAYCTRL_KEYOFF);
	REG_WRITE_4(AICA_PLAYCTRL_REG(rch), AICA_PLAYCTRL_KEYOFF);

	ratepitch = rate2reg(unitwork[unit].rate);

	/* setup left */
	REG_WRITE_4(AICA_LOOP_START_REG(lch), 0);
	REG_WRITE_4(AICA_LOOP_END_REG(lch), unitwork[unit].nsample * unitwork[unit].nblk);
	REG_WRITE_4(AICA_RATE_REG(lch), ratepitch);
	REG_WRITE_4(AICA_VOLPAN_REG(lch),
	    AICA_VOLPAN_VOL(15) |	/* -0dB */
	    AICA_VOLPAN_PAN(0x1f));	/* left PAN */

	/* setup right */
	REG_WRITE_4(AICA_LOOP_START_REG(rch), 0);
	REG_WRITE_4(AICA_LOOP_END_REG(rch), unitwork[unit].nsample * unitwork[unit].nblk);
	REG_WRITE_4(AICA_RATE_REG(rch), ratepitch);
	REG_WRITE_4(AICA_VOLPAN_REG(rch),
	    AICA_VOLPAN_VOL(15) |	/* -0db */
	    AICA_VOLPAN_PAN(0x0f));	/* left PAN */

	switch (unitwork[unit].precision) {
	case 4:
		mode = AICA_PLAYCTRL_LONGADPCM;
		break;
	case 8:
		mode = AICA_PLAYCTRL_PCM8BIT;
		break;
	case 16:
	default:
		mode = AICA_PLAYCTRL_PCM16BIT;
		break;
	}
	mode |= AICA_PLAYCTRL_KEYON | AICA_PLAYCTRL_LOOP;

	switch (unitwork[unit].channel) {
	case 1:
		REG_WRITE_4(AICA_VOLPAN_REG(lch),
		    AICA_VOLPAN_VOL(15) |	/* -0db */
		    AICA_VOLPAN_PAN(0x00));	/* middle PAN */

		REG_WRITE_4(AICA_ADDRLO_REG(lch),
		    AICA_ADDRLO_ADDR(AICA_DMABUF_LEFT(unit)));

		REG_WRITE_4(AICA_PLAYCTRL_REG(lch),
		    mode | AICA_PLAYCTRL_ADDRHI(AICA_DMABUF_LEFT(unit)));
		break;
	case 2:
		REG_WRITE_4(AICA_ADDRLO_REG(lch),
		    AICA_ADDRLO_ADDR(AICA_DMABUF_LEFT(unit)));
		REG_WRITE_4(AICA_ADDRLO_REG(rch),
		    AICA_ADDRLO_ADDR(AICA_DMABUF_RIGHT(unit)));

		REG_WRITE_4(AICA_PLAYCTRL_REG(lch),
		    mode | AICA_PLAYCTRL_ADDRHI(AICA_DMABUF_LEFT(unit)));
		REG_WRITE_4(AICA_PLAYCTRL_REG(rch),
		    mode | AICA_PLAYCTRL_ADDRHI(AICA_DMABUF_RIGHT(unit)));
		break;
	}

	unitwork[unit].curpos = 0;
	unitwork[unit].state = UNIT_STATE_PLAYING;
}

void
aica_stop(int unit)
{
	int lch, rch;

	lch = unit * 2 + 0;
	rch = unit * 2 + 1;

	bzero32(AICA_DMABUF_LEFT(unit), AICA_DMABUF_SIZE);
	bzero32(AICA_DMABUF_RIGHT(unit), AICA_DMABUF_SIZE);

	CH_WRITE_4(lch, 0x00, AICA_PLAYCTRL_KEYOFF);
	CH_WRITE_4(rch, 0x00, AICA_PLAYCTRL_KEYOFF);
}

static void
fill_silent(int unit, uint32_t pos, uint32_t blksize)
{
	void *p;

	p = (void*)AICA_DMABUF_LEFT(unit);
	bzero32((char *)p + pos, blksize);

	if (unitwork[unit].channel > 1) {
		p = (void *)AICA_DMABUF_RIGHT(unit);
		bzero32((char *)p + pos, blksize);
	}
}



void
aica_main(void)
{
	/* command queue */
	struct cmdqueue cmdq_to_sh4;
	struct cmdqueue cmdq_from_sh4;
	struct cmd *cmd, tmpcmd;
	int nloop, unit, need_to_intr;

	cmdq_to_sh4.address = AICA_ARMtoSH4_CMDQBUF;
	cmdq_to_sh4.size    = AICA_ARMtoSH4_CMDQSIZE,
	cmdq_to_sh4.index   = 0;
	cmdq_to_sh4.serial  = 0;
	cmdq_from_sh4.address = AICA_SH4toARM_CMDQBUF;
	cmdq_from_sh4.size    = AICA_SH4toARM_CMDQSIZE,
	cmdq_from_sh4.index   = 0;
	cmdq_from_sh4.serial  = 0;

	aica_init();

	bzero32(unitwork, sizeof(unitwork));

	/* Enable Interrupt to SH4 */
	REG_WRITE_4(AICA_INTEN_REG, AICA_INT_CPU);
	REG_WRITE_4(AICA_INTLVL2_REG, 8);

	need_to_intr = 0;
	for (nloop = 0xffffffff; ; nloop++) {
		*(volatile uint32_t *)AICA_HEARTBEAT = nloop;
		*(volatile uint32_t *)AICA_DEBUGPORT1 = cmdq_to_sh4.index;
		*(volatile uint32_t *)AICA_DEBUGPORT2 = cmdq_to_sh4.serial;

		while ((cmd = aica_readcmd(&cmdq_from_sh4)) != NULL) {
			unit = AICA_COMMAND_UNIT(cmd->command);
			switch (AICA_COMMAND_CMD(cmd->command)) {
			case AICA_CMD_PLAY:
				aica_play(unit);

				aica_printf("PLAY: ch=%d, rate=%d, prec=%d, nblk=%d, nsample=%d\n",
				    unitwork[unit].channel,
				    unitwork[unit].rate,
				    unitwork[unit].precision,
				    unitwork[unit].nblk,
				    unitwork[unit].nsample);
				break;

			case AICA_CMD_STOP:
				unitwork[unit].state = UNIT_STATE_STOPPING;
				unitwork[unit].leftblk = unitwork[unit].nblk;
				aica_printf("STOP: ch=%d, rate=%d, prec=%d, nblk=%d, nsample=%d\n",
				    unitwork[unit].channel,
				    unitwork[unit].rate,
				    unitwork[unit].precision,
				    unitwork[unit].nblk,
				    unitwork[unit].nsample);
				break;
			case AICA_CMD_CH_RATE:
				unitwork[unit].channel = AICA_PARAM_HI(cmd->param);
				unitwork[unit].rate = AICA_PARAM_LO(cmd->param);

				aica_printf("RECV: ch=%d, rate=%d\n",
				    AICA_PARAM_HI(cmd->param),
				    AICA_PARAM_LO(cmd->param));
				break;
			case AICA_CMD_PREC_BLKSIZE:
				unitwork[unit].precision = AICA_PARAM_HI(cmd->param);
				unitwork[unit].blksize = AICA_PARAM_LO(cmd->param);

				aica_printf("RECV: precision=%d, blksize=%d\n",
				    AICA_PARAM_HI(cmd->param),
				    AICA_PARAM_LO(cmd->param));
				break;
			case AICA_CMD_NBLK_NSMPL:
				unitwork[unit].nblk = AICA_PARAM_HI(cmd->param);
				unitwork[unit].nsample = AICA_PARAM_LO(cmd->param);

				aica_printf("RECV: nblk=%d, nsample=%d\n",
				    AICA_PARAM_HI(cmd->param),
				    AICA_PARAM_LO(cmd->param));
				break;
			default:
				break;
			}
		}

		for (unit = 0; unit < MAXUNIT; unit++) {
			switch (unitwork[unit].state) {
			case UNIT_STATE_IDLE:
				break;
			case UNIT_STATE_PLAYING:
				{
					uint32_t pos, ndone;

					REG_WRITE_4(AICA_CHANSEL_REG, AICA_CHANSEL_CH(unit * 2 + 0));
					pos = REG_READ_4(AICA_CHANPOS_REG) & 0xffff;

					for (ndone = 0;
					    (pos >= (unitwork[unit].curpos + unitwork[unit].nsample)) ||
					    (unitwork[unit].curpos > pos); ) {

						ndone++;
						unitwork[unit].curpos += unitwork[unit].nsample;
						if (unitwork[unit].curpos >= (unitwork[unit].nsample * unitwork[unit].nblk))
							unitwork[unit].curpos = 0;
					}
					if (ndone) {
						tmpcmd.command = AICA_COMMAND_PACK(AICA_CMD_FORWARD, unit);
						tmpcmd.param = ndone;
						aica_sendcmd(&cmdq_to_sh4, &tmpcmd);
						need_to_intr = 1;
					}
				}
				break;
			case UNIT_STATE_STOPPING:
				{
					uint32_t pos, ndone;

					REG_WRITE_4(AICA_CHANSEL_REG, AICA_CHANSEL_CH(unit * 2 + 0));
					pos = REG_READ_4(AICA_CHANPOS_REG) & 0xffff;

					for (ndone = 0;
					    (pos >= (unitwork[unit].curpos + unitwork[unit].nsample)) ||
					    (unitwork[unit].curpos > pos); ) {

						ndone++;

						fill_silent(unit, unitwork[unit].curpos, unitwork[unit].blksize);

						aica_printf("SILENCE: unit=%d curpos=0x%08x blksize=%d\n",
						    unit, unitwork[unit].curpos, unitwork[unit].blksize);

						unitwork[unit].curpos += unitwork[unit].nsample;
						if (unitwork[unit].curpos >= (unitwork[unit].nsample * unitwork[unit].nblk))
							unitwork[unit].curpos = 0;
					}
					if (ndone) {
						aica_printf("SILENCE: ndone=%d, leftblk=%d\n", ndone, unitwork[unit].leftblk);

						unitwork[unit].leftblk -= ndone;
						if (unitwork[unit].leftblk <= 0) {
							unitwork[unit].state = UNIT_STATE_IDLE;
							aica_stop(unit);
						}
					}
					break;
				}
			}
		}

#ifdef AICA_DEBUG
		if ((nloop & 0x1ffff) == 0)
			aica_printf("NLOOP=%d\n", nloop);
#endif

#ifdef AICA_PRINTF
		if (aica_msgbuf_update()) {
			tmpcmd.command = AICA_CMD_MSGBUF;
			tmpcmd.param = nloop;
			aica_sendcmd(&cmdq_to_sh4, &tmpcmd);
			need_to_intr = 1;
		}
#endif

		if (need_to_intr) {
			need_to_intr = 0;
			REG_WRITE_4(AICA_INTSET_REG, AICA_INT_CPU);
		}
	}

#if 0
//	volatile aica_cmd_t *aicacmd = (volatile aica_cmd_t *)AICA_ARM_CMD;
//	int play_state;
//	unsigned int loopend = 0,loophalf = 0;
//	unsigned int blksize = 0, ratepitch;
//	uint32_t cmd, serial;
//
//	play_state = 0;
//	serial = aicacmd->serial = 0;
//
//	for (;;) {
//		if (serial != aicacmd->serial) {
//			serial = aicacmd->serial;
//			cmd = aicacmd->command;
//			aicacmd->command = AICA_COMMAND_NOP;
//		} else {
//			cmd = AICA_COMMAND_NOP;
//		}
//
//		switch (cmd) {
//		case AICA_COMMAND_NOP:
//			/*
//			 * AICA_COMMAND_NOP - Idle process
//			 */
//			switch (play_state) {
//			case 0: /* not playing */
//				break;
//			case 1: /* first half */
//				if (in_second_half(loophalf)) {
//					/* Send INT to SH4 */
//					REG_WRITE_4(0x28b8, 0x0020);
//					play_state = 2;
//				}
//				break;
//			case 2: /* second half */
//				if (in_first_half(loophalf)) {
//					/* Send INT to SH4 */
//					REG_WRITE_4(0x28b8, 0x0020);
//					play_state = 1;
//				}
//				break;
//			case 3:
//				if (in_second_half(loophalf)) {
//					aica_stop();
//					play_state = 0;
//				}
//				break;
//			case 4:
//				if (in_first_half(loophalf)) {
//					aica_stop();
//					play_state = 0;
//				}
//				break;
//			}
//			break;
//
//		case AICA_COMMAND_PLAY:
//			blksize = aicacmd->blocksize;
//
//			CH_WRITE_4(0, 0x00, 0x8000);
//			CH_WRITE_4(1, 0x00, 0x8000);
//
//			switch (aicacmd->precision) {
//			case 16:
//				loopend = blksize;
//				break;
//			case 8:
//				loopend = blksize * 2;
//				break;
//			case 4:
//				loopend = blksize * 4;
//				break;
//			}
//			loophalf = loopend / 2;
//
//			ratepitch = rate2reg(aicacmd->rate);
//
//			/* setup left */
//			CH_WRITE_4(0, 0x08, 0);		/* loop start */
//			CH_WRITE_4(0, 0x0c, loopend);	/* loop end */
//			CH_WRITE_4(0, 0x18, ratepitch);	/* SamplingRate */
//			CH_WRITE_4(0, 0x24, 0x0f1f);	/* volume MAX,
//							   right PAN */
//
//			/* setup right */
//			CH_WRITE_4(1, 0x08,0);		/* loop start */
//			CH_WRITE_4(1, 0x0c, loopend);	/* loop end */
//			CH_WRITE_4(1, 0x18, ratepitch);	/* SamplingRate */
//			CH_WRITE_4(1, 0x24, 0x0f0f);	/* volume MAX,
//							   right PAN */
//
//			{
//				uint32_t mode, lparam, rparam;
//
//				if (aicacmd->precision == 4)
//					mode = 3 << 7;	/* 4bit ADPCM */
//				else if (aicacmd->precision == 8)
//					mode = 1 << 7;	/* 8bit */
//				else
//					mode = 0;	/* 16bit */
//
//				switch (aicacmd->channel) {
//				case 2:
//					CH_WRITE_4(0, 0x04,
//					    AICA_DMABUF_LEFT & 0xffff);
//					CH_WRITE_4(1, 0x04,
//					    AICA_DMABUF_RIGHT & 0xffff);
//					lparam = 0xc000 /*PLAY*/ |
//					    0x0200 /*LOOP*/ | mode |
//					    (AICA_DMABUF_LEFT >> 16);
//					rparam = 0xc000 /*PLAY*/ |
//					    0x0200 /*LOOP*/ | mode |
//					    (AICA_DMABUF_RIGHT >> 16);
//					CH_WRITE_4(0, 0x00, lparam);
//					CH_WRITE_4(1, 0x00, rparam);
//					break;
//				case 1:
//					CH_WRITE_1(0, 0x24, 0);	/* middle
//								   balance */
//					CH_WRITE_4(0, 0x04,
//					    AICA_DMABUF_LEFT & 0xffff);
//					CH_WRITE_4(0, 0x00, 0xc000 /*PLAY*/ |
//					    0x0200 /*LOOP*/ | mode |
//					    (AICA_DMABUF_LEFT >> 16));
//					break;
//				}
//			}
//			play_state = 1;
//			break;
//
//		case AICA_COMMAND_STOP:
//			switch (play_state) {
//			case 1:
//				bzero32((void *)(AICA_DMABUF_LEFT + blksize),
//				    blksize);
//				bzero32((void *)(AICA_DMABUF_RIGHT + blksize),
//				    blksize);
//				play_state = 3;
//				break;
//			case 2:
//				bzero32((void *)AICA_DMABUF_LEFT, blksize);
//				bzero32((void *)AICA_DMABUF_RIGHT, blksize);
//				play_state = 4;
//				break;
//			default:
//				aica_stop();
//				play_state = 0;
//				break;
//			}
//			break;
//
//		case AICA_COMMAND_INIT:
//			aica_stop();
//			play_state = 0;
//			break;
//
//		case AICA_COMMAND_MVOL:
//			REG_WRITE_4(0x2800, L256TO16(aicacmd->l_param));
//			break;
//
//		case AICA_COMMAND_VOL:
//			CH_WRITE_1(0, 0x29, 0xff - (aicacmd->l_param & 0xff));
//			CH_WRITE_1(1, 0x29, 0xff - (aicacmd->r_param & 0xff));
//			break;
//
//		}
//	}
#endif
}
